import functools
import inspect
import sys
from functools import lru_cache, wraps
from typing import Any, Callable, Dict, Mapping, Union

if sys.version_info < (3, 9):
    from typing_extensions import Annotated, get_args, get_origin, get_type_hints
else:
    from typing import Annotated, get_type_hints, get_origin, get_args

from di.exceptions import WiringError


def cached_accept_callable_class(
    maxsize: int,
) -> Callable[[Callable[..., bool]], Callable[..., bool]]:
    def wrapper(func: Callable[..., bool]) -> Callable[..., bool]:
        func = lru_cache(maxsize=maxsize)(func)

        @wraps(func)
        def inner(call: Any) -> bool:
            if not callable(call):
                return False
            if inspect.isclass(call):
                return False
            if isinstance(call, functools.partial):
                call = call.func
            if func(call):
                return True
            _call = getattr(call, "__call__", None)
            if _call is None:
                return False
            return func(_call)

        return inner

    return wrapper


@cached_accept_callable_class(maxsize=2 ** 10)
def is_coroutine_callable(call: Callable[..., Any]) -> bool:
    return inspect.iscoroutinefunction(call)


@cached_accept_callable_class(maxsize=2 ** 10)
def is_async_gen_callable(call: Callable[..., Any]) -> bool:
    return inspect.isasyncgenfunction(call)


@cached_accept_callable_class(maxsize=2 ** 10)
def is_gen_callable(call: Any) -> bool:
    return inspect.isgeneratorfunction(call)


@lru_cache(maxsize=2 ** 10)
def get_annotations(call: Callable[..., Any]) -> Dict[str, Any]:
    types_from: Callable[..., Any]
    if not (
        inspect.isclass(call) or inspect.isfunction(call) or inspect.ismethod(call)
    ) and hasattr(call, "__call__"):
        # callable class
        types_from = call.__call__  # type: ignore[misc,operator] # accessing __init__ directly
    else:
        # method
        types_from = call
    hints = get_type_hints(types_from, include_extras=True)
    # for no apparent reason, Annotated[Optional[T]] comes back as Optional[Annotated[Optional[T]]]
    # so remove the outer Optional if this is the case
    for param_name, hint in hints.items():
        args = get_args(hint)
        if get_origin(hint) is Union and get_origin(next(iter(args))) is Annotated:
            hints[param_name] = next(iter(args))
    return hints


@lru_cache(maxsize=2 ** 10)
def get_parameters(call: Callable[..., Any]) -> Dict[str, inspect.Parameter]:
    params: Mapping[str, inspect.Parameter]
    if inspect.isclass(call) and (call.__new__ is not object.__new__):  # type: ignore[comparison-overlap]
        # classes overriding __new__, including some generic metaclasses, result in __new__ getting read
        # instead of __init__
        params = inspect.signature(call.__init__).parameters  # type: ignore[misc] # accessing __init__ directly
        params = dict(params)
        params.pop(next(iter(params.keys())))  # first parameter to __init__ is self
    else:
        params = inspect.signature(call).parameters
    annotations = get_annotations(call)
    processed_params: Dict[str, inspect.Parameter] = {}
    for param_name, param in params.items():
        param = param.replace(annotation=annotations.get(param_name, param.annotation))
        processed_params[param_name] = param
    return processed_params


def infer_call_from_annotation(parameter: inspect.Parameter) -> Callable[..., Any]:
    if not callable(parameter.annotation):
        raise WiringError(
            f"Annotation for {parameter.name} is not a callable class or function and so we cannot autowire it."
            " You must explicity provide a default value implementing the DependantBase"
        )
    return parameter.annotation  # type: ignore[no-any-return]



import sys
from typing import Any, Collection, ContextManager, Mapping, Optional, TypeVar, Union

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

from di._utils.types import FusedContextManager
from di.api.dependencies import DependantBase
from di.api.executor import AsyncExecutor, SyncExecutor
from di.api.providers import DependencyProvider, DependencyProviderType
from di.api.scopes import Scope
from di.api.solved import SolvedDependant

ContainerType = TypeVar("ContainerType")

DependencyType = TypeVar("DependencyType")


class ContainerProtocol(Protocol):
    @property
    def scopes(self) -> Collection[Scope]:
        ...

    def bind(
        self,
        provider: Union[
            DependantBase[DependencyType], DependencyProviderType[DependencyType]
        ],
        dependency: DependencyProviderType[DependencyType],
    ) -> ContextManager[None]:
        """Bind a new dependency provider for a given dependency.

        This can be used as a function (for a permanent bind, cleared when `scope` is exited)
        or as a context manager (the bind will be cleared when the context manager exits).

        Binds are only identified by the identity of the callable and do not take into account
        the scope or any other data from the dependency they are replacing.
        """
        ...

    def solve(
        self,
        dependency: DependantBase[DependencyType],
    ) -> SolvedDependant[DependencyType]:
        """Solve a dependency.

        Returns a SolvedDependant that can be executed to get the dependency's value.
        """
        ...

    def enter_scope(
        self: ContainerType, scope: Scope
    ) -> FusedContextManager[ContainerType]:
        """Enter a scope and get back a new BaseContainer in that scope"""
        ...

    def execute_sync(
        self,
        solved: SolvedDependant[DependencyType],
        *,
        executor: Optional[SyncExecutor] = None,
        values: Optional[Mapping[DependencyProvider, Any]] = None,
    ) -> DependencyType:
        """Execute an already solved dependency.

        This method may or may not be able to handle async dependencies,
        this is up to the executor passed in.
        """
        ...

    async def execute_async(
        self,
        solved: SolvedDependant[DependencyType],
        *,
        executor: Optional[AsyncExecutor] = None,
        values: Optional[Mapping[DependencyProvider, Any]] = None,
    ) -> DependencyType:
        """Execute an already solved dependency.

        This method can always handle sync and async dependencies.
        """
        ...



from __future__ import annotations

from collections import deque
from contextlib import AsyncExitStack, ExitStack
from typing import (
    AbstractSet,
    Any,
    Deque,
    Dict,
    Iterable,
    Mapping,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Union,
)

from graphlib2 import TopologicalSorter

from di._utils.scope_validation import validate_scopes
from di._utils.task import AsyncTask, ExecutionState, SyncTask, gather_new_tasks
from di.api.dependencies import DependantBase
from di.api.executor import State as ExecutorState
from di.api.executor import Task as ExecutorTask
from di.api.providers import DependencyProvider
from di.api.scopes import Scope
from di.api.solved import SolvedDependant

Dependency = Any

Task = Union[AsyncTask, SyncTask]

TaskCacheDeque = Deque[Tuple[Task, Scope]]

Results = Dict[Task, Any]

TaskDependencyCounts = Dict[Task, int]


class SolvedDependantCache(NamedTuple):
    """Private data that the Container attaches to SolvedDependant"""

    root_task: Task
    topological_sorter: TopologicalSorter[Task]
    cacheable_tasks: AbstractSet[Task]
    cached_tasks: Iterable[Tuple[Task, Scope]]
    validated_scopes: Set[Tuple[Scope, ...]]
    call_map: Mapping[DependencyProvider, Iterable[Task]]


def plan_execution(
    stacks: Mapping[Scope, Union[AsyncExitStack, ExitStack]],
    cached_values: Mapping[DependantBase[Any], Any],
    solved: SolvedDependant[Any],
    *,
    values: Optional[Mapping[DependencyProvider, Any]] = None,
) -> Tuple[
    Dict[Task, Any],
    Iterable[Optional[ExecutorTask]],
    ExecutorState,
    Iterable[Tuple[Task, Scope]],
    Task,
]:
    """Re-use or create an ExecutionPlan"""
    # This function is a hot loop
    # It is run for every execution, and even with the cache it can be a bottleneck
    # This would be the best place to Cythonize or otherwise take more drastic measures

    user_values = values or {}

    solved_dependency_cache: SolvedDependantCache = solved.container_cache

    scopes = tuple(stacks.keys())
    if scopes not in solved_dependency_cache.validated_scopes:
        validate_scopes(scopes, solved.dag)
        solved_dependency_cache.validated_scopes.add(scopes)

    results: Results = {}
    call_map = solved_dependency_cache.call_map
    for call in user_values.keys() & call_map.keys():
        value = user_values[call]
        for task_id in call_map[call]:
            results[task_id] = value

    to_cache: TaskCacheDeque = deque()
    cacheable_tasks = solved_dependency_cache.cacheable_tasks
    add_to_cache = to_cache.append

    for task, scope in solved_dependency_cache.cached_tasks:
        if task.dependant in cached_values:
            results[task] = cached_values[task.dependant]
        elif task in cacheable_tasks:
            add_to_cache((task, scope))

    ts = solved_dependency_cache.topological_sorter.copy()

    execution_state = ExecutionState(
        stacks=stacks,
        results=results,
        toplogical_sorter=ts,
    )

    return (
        results,
        gather_new_tasks(execution_state),
        ExecutorState(execution_state),
        to_cache,
        solved_dependency_cache.root_task,
    )
